# from openai import OpenAI, OpenAIError, BadRequestError
# from openai.types.beta import Assistant, Thread
# from openai.types.beta.threads import Message, Run, Text
# from openai.types.beta.threads.run import LastError
# from openai.types.beta.threads.runs import RunStep
# from openai.types.beta.threads.runs.message_creation_step_details import (
#     MessageCreationStepDetails,
# )
# from openai.types.beta.threads.runs.tool_calls_step_details import ToolCallsStepDetails
# from openai.types.beta.threads.file_citation_annotation import FileCitationAnnotation
# from openai.types.beta.threads.text_content_block import TextContentBlock
# from openai.types.beta.threads.image_file_content_block import ImageFileContentBlock
# from openai.types.beta.threads.file_path_annotation import FilePathAnnotation
# from openai.types.beta.threads.runs.function_tool_call import FunctionToolCall, Function
# from openai.types.beta.threads.runs.code_interpreter_tool_call import (
#     CodeInterpreterToolCall,
# )
# from openai.types.beta.code_interpreter_tool import CodeInterpreterTool
# from openai.types.beta.function_tool import FunctionTool

# from tiktoken import get_encoding, encoding_for_model
from typing import Optional, Union, Any, Literal, Mapping
from json import loads as json_loads, JSONDecodeError



from openai import OpenAI
from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.chat.chat_completion_message_tool_call import ChatCompletionMessageToolCall

from ._types import Message, Tool, ToolCall
from ._convert_types import stringify_content
from .baseaiclientwrapper import BaseAIClientWrapper

class OpenAIWrapper(BaseAIClientWrapper):
    client: OpenAI
    default_model = "gpt-4o-mini"

    def import_client(self):
        from openai import OpenAI
        return OpenAI
    
    
    def prep_input_message(self, message: Message) -> dict:
        message_dict = {
            "role": message.role,
            "content": message.content,                
        }
        if message.tool_calls:
            if message.role == "tool":
                tool_call = message.tool_calls[0]
                message_dict["tool_call_id"] = tool_call.tool_call_id
            else:
                message_dict["tool_calls"] = [
                    self.prep_input_tool_call(tool_call) for tool_call in message.tool_calls
                ]

        if message.images:
            # TODO: Implement image handling
            message_dict["images"] = message.images

        return message_dict
    
    def prep_input_messages_and_system_prompt(self, 
                                              messages: list[Message], 
                                              system_prompt_arg: Optional[str] = None
                                              ) -> tuple[list, Optional[str]]:
        if system_prompt_arg:
            system_prompt = system_prompt_arg
            if messages and messages[0].role == "system":
                messages[0].content = system_prompt
            else:
                messages.insert(0, Message(role="system", content=system_prompt))
        elif messages and messages[0].role == "system":
            system_prompt = messages[0].content
        else:
            system_prompt = None

        client_messages = [self.prep_input_message(message) for message in messages]
        return client_messages, system_prompt

    
    def prep_input_tool(self, tool: Tool) -> dict:
        return tool.to_dict()
    
    def prep_input_tool_choice(self, tool_choice: str) -> Union[str, dict]:
        if tool_choice in ("auto", "required", "none"):
            return tool_choice

        tool_type = "function" # Currently only function tools are supported See: https://platform.openai.com/docs/api-reference/chat/create#chat-create-tool_choice
        return {"type": tool_type, tool_type: {"name": tool_choice}}

    def prep_input_tool_call(self, tool_call: ToolCall) -> dict:
        return {
            "id": tool_call.tool_call_id,
            "type": tool_call.type,
            tool_call.type: {
                "name": tool_call.tool_name,
                "arguments": stringify_content(tool_call.arguments),
            }
        }


    def split_tool_call_outputs_into_messages(self, tool_calls: list[ToolCall]) -> list[Message]:        
        return [
            Message(role="tool", content=stringify_content(tool_call.output), tool_calls=[tool_call])
            for tool_call in tool_calls
        ]
    
    # def prep_input_tool_call_response(self, tool_call: ToolCall, tool_response: Any) -> dict:
    #     return {
    #         "role": "tool",
    #         "tool_call_id": tool_call.tool_call_id,
    #         "name": tool_call.tool_name,
    #         "content": tool_response,
    #     }

    def prep_input_response_format(self, response_format: Union[str, dict]) -> Union[str, dict]:
        simple_response_formats = ("json", "json_object", "text")
        
        if isinstance(response_format, dict) and (response_type := response_format.get("type")):
            if response_type == "json_schema" and (schema := response_format.get("json_schema")):
                # TODO handle json_schema
                # schema = handle_json_schema(schema)
                return {"type": response_type, response_type: schema}
        else:
            response_type = response_format    
        
        if response_type in ("json", "json_object", "text"):
            return {"type": response_type}
        
        raise ValueError(f"Invalid response_format: {response_format}")
        

    def extract_output_tool_call(self, tool_call: ChatCompletionMessageToolCall) -> ToolCall:
        return ToolCall(
            tool_call_id=tool_call.id,
            tool_name=tool_call.function.name,
            arguments=json_loads(tool_call.function.arguments),
        )    
    
    
    # def extract_output_message(self, response: ChatCompletion) -> Message:
    #     message = response.choices[0].message
    #     images = None # TODO: Implement image extraction
    #     tool_calls = [self.extract_output_tool_call(tool_call) for tool_call in message.tool_calls] if message.tool_calls else None
    #     return Message(
    #         role=message.role,
    #         content=message.content,
    #         images=images,
    #         tool_calls=tool_calls,
    #         response_object=response,
    #     )

    def extract_std_and_client_messages(self, response: ChatCompletion) -> tuple[Message, ChatCompletionMessage]:
        client_message = response.choices[0].message
        images = None # TODO: Implement image extraction
        tool_calls = [self.extract_output_tool_call(tool_call) for tool_call in client_message.tool_calls] if client_message.tool_calls else None
        std_message = Message(
            role=client_message.role,
            content=client_message.content,
            images=images,
            tool_calls=tool_calls,
            response_object=response,
        )   
        return std_message, client_message


    def list_models(self) -> list[str]:
        return [model.id for model in self.client.models.list()]

    def chat(
            self,
            messages: list[dict],     
            model: str = default_model,
            system_prompt: Optional[str] = None,                    
            tools: Optional[list[dict]] = None,
            tool_choice: Optional[Union[Literal["auto", "required", "none"], dict]] = None,
            response_format: Optional[str] = None,
            **kwargs
            ) -> ChatCompletion:

            if tool_choice and not tools:
                tool_choice = None

            response = self.client.chat.completions.create(
                messages=messages, 
                model=model,
                tools=tools,
                tool_choice=tool_choice,
                response_format=response_format,
                **kwargs
            )
            return response




import pytest
from unifai import UnifAIClient

from os import getenv
from dotenv import load_dotenv

load_dotenv()
ANTHROPIC_API_KEY = getenv("ANTHROPIC_API_KEY")
OPENAI_API_KEY = getenv("OPENAI_API_KEY")

def base_test_all_providers(func):
    return pytest.mark.parametrize("provider, client_kwargs, func_kwargs", [
    (
        "anthropic", 
        {"api_key": ANTHROPIC_API_KEY},
        {}
        ),
    (
        "openai", 
        {"api_key": OPENAI_API_KEY},
        {}
    ),
    (
        "ollama", 
        {"host": "http://librem-2.local:11434"},
        {"keep_alive": "10m", 'model': 'llama3.1-8b-num_ctx-8192:latest'}
    ),
    ])(func)




import pytest
from unifai import UnifAIClient, AIProvider
from unifai._types import Message, Tool
from basetest import base_test_all_providers

# TOOLS AND TOOL CALLABLES
TOOLS = {
    "get_current_weather": 
    {
        "type": "function",
        "function": {
            "name": "get_current_weather",
            "description": "Get the current weather in a given location",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The city and state, e.g. San Francisco, CA",
                    },
                    "unit": {
                        "type": "string", 
                        "description": "The unit of temperature to return. Infer the unit from the location if not provided.",
                        "enum": ["celsius", "fahrenheit"]
                    },
                },
                "required": ["location", "unit"],
            },
        }
    },
    "return_weather_messages":
    {
        "type": "function",
        "function": {
            "name": "return_weather_messages",
            "description": "Return a message about the current weather for one or more locations",
            "parameters": {
                "type": "object",
                "properties": {
                    "messages": {
                        "type": "array",
                        "description": "The messages to return about the weather",
                        "items":{
                            "type": "object",
                            "properties": {
                                "location": {
                                    "type": "string",
                                    "description": "The city and state, e.g. San Francisco, CA",
                                },
                                "message": {
                                    "type": "string",
                                    "description": "The message for the weather in the location",
                                },
                            },
                            "required": ["location", "message"],
                        },
                    },
                },
                "required": ["messages"],
            }
        }
    },

}

def get_current_weather(location: str, unit: str = "fahrenheit") -> dict:
    location = location.lower()
    if 'san francisco' in location:
        degrees = 69
        condition = 'sunny'
    elif 'tokyo' in location:
        degrees = 50
        condition = 'cloudy'
    elif 'paris' in location:
        degrees = 40
        condition = 'rainy'
    else:
        degrees = 100
        condition = 'hot'
    if unit == 'celsius':
        degrees = (degrees - 32) * 5/9
        unit = 'C'
    else:
        unit = 'F'
    return {'condition': condition, 'degrees': degrees, 'unit': unit}












@base_test_all_providers
def test_chat_simple(provider: AIProvider, client_kwargs: dict, func_kwargs: dict):

    ai = UnifAIClient({provider: client_kwargs})
    ai.init_client(provider, **client_kwargs)
    messages = ai.chat(
        messages=[{"role": "user", "content": "Hello, how are you?"}],
        provider=provider,
        **func_kwargs
    )
    assert messages
    assert isinstance(messages, list)

    for message in messages:
        assert isinstance(message, Message)
        assert message.content
        print(f'{message.role}: {message.content}')
        if message.tool_calls:
            for tool_call in message.tool_calls:
                print(f'Tool Call: {tool_call.tool_name}')
                print(tool_call.arguments)
    print()



@base_test_all_providers
@pytest.mark.parametrize("messages, tools, tool_callables", [
    (
        [
            {"role": "system", "content": "Your role is use the available tools to answer questions like a cartoon Pirate"},            
            {"role": "user", "content": "What's the weather like in San Francisco, Tokyo, and Paris? Respond in Fahrenheit."},
        ],
        [
            TOOLS["get_current_weather"]
        ],
        {
            "get_current_weather": get_current_weather
        }
    ),

])
def test_chat_tools_simple(
    # ai: UnifAIClient, 
    provider: AIProvider, 
    client_kwargs: dict,
    func_kwargs: dict,
    messages: list,
    tools: list,
    tool_callables: dict
    ):

    ai = UnifAIClient(
        {provider: client_kwargs},
        tool_callables=tool_callables
    )
    ai.init_client(provider, **client_kwargs)
    messages = ai.chat(
        messages=messages,
        provider=provider,
        tools=tools,
        **func_kwargs
    )
    assert messages
    assert isinstance(messages, list)

    for message in messages:
        print(f'{message.role}:\n{message.content or message.tool_calls}\n')
        assert isinstance(message, Message)
        assert message.content or message.tool_calls
    print()
    


@base_test_all_providers
@pytest.mark.parametrize("messages, tools, tool_callables", [
    (
        [
            {"role": "system", "content": "Your role is use the available tools to answer questions like a cartoon Pirate"},            
            {"role": "user", "content": "What's the weather like in San Francisco, Tokyo, and Paris? Respond in Fahrenheit."},
        ],
        [
            TOOLS["get_current_weather"]
        ],
        {
            "get_current_weather": get_current_weather
        }
    ),
])
def test_chat_return_on(
    # ai: UnifAIClient, 
    provider: AIProvider, 
    client_kwargs: dict,
    func_kwargs: dict,
    messages: list,
    tools: list,
    tool_callables: dict, 
    ):

    ai = UnifAIClient(
        {provider: client_kwargs},
        tool_callables=tool_callables
    )
    ai.init_client(provider, **client_kwargs)

    return_ons = ["content"]
    if tool_names := [tool["function"]["name"] for tool in tools]:
        return_ons.append("message")
        return_ons.append("tool_call")
        return_ons.append(tool_names[0])
        return_ons.append(tool_names)
    

    for return_on in return_ons:
        new_messages = ai.chat(
            messages=messages,
            provider=provider,
            tools=tools,
            return_on=return_on,
            **func_kwargs
        )

        assert new_messages
        last_message = new_messages[-1]
        assert isinstance(last_message, Message)

        if return_on == "content":
            assert last_message.content
            assert not last_message.tool_calls

        elif return_on == "message":
            assert last_message.content or last_message.tool_calls

        elif return_on == "tool_call":
            assert last_message.tool_calls
            # assert not last_message.content

        elif tool_names and return_on == tool_names[0]:
            assert last_message.tool_calls
            # assert not last_message.content
            assert last_message.tool_calls[0].tool_name == tool_names[0]

        elif tool_names and return_on == tool_names:
            assert last_message.tool_calls
            # assert not last_message.content
            assert last_message.tool_calls[0].tool_name in tool_names            

        for message in new_messages:
            print(f'\n{message.role}:\n{message.content or message.tool_calls}\n')            
        print()




@base_test_all_providers
@pytest.mark.parametrize("messages, tools, tool_callables, tool_choice", [
    (
        [
            {"role": "system", "content": "Your role is use the available tools to answer questions like a cartoon Pirate"},            
            {"role": "user", "content": "What's the weather like in San Francisco, Tokyo, and Paris? Respond in Fahrenheit."},
        ],
        [
            TOOLS["get_current_weather"]
        ],
        {
            "get_current_weather": get_current_weather
        }, 
        "get_current_weather"
    ),
])
def test_chat_enforce_tool_choice(
    # ai: UnifAIClient, 
    provider: AIProvider, 
    client_kwargs: dict,
    func_kwargs: dict,
    messages: list,
    tools: list,
    tool_callables: dict, 
    tool_choice: str
    ):

    ai = UnifAIClient(
        {provider: client_kwargs},
        tool_callables=tool_callables
    )
    ai.init_client(provider, **client_kwargs)

    for _ in range(1):
        new_messages = ai.chat(
            messages=messages,
            provider=provider,
            tools=tools,
            tool_choice=tool_choice,
            return_on="message",
            enforce_tool_choice=True,
            **func_kwargs
        )
        
        assert new_messages
        assert isinstance(new_messages, list)

        last_message = new_messages[-1]
        assert isinstance(last_message, Message)

        if tool_choice == 'auto':
            assert last_message.content or last_message.tool_calls
        elif tool_choice == 'required':
            assert last_message.tool_calls
            assert not last_message.content
        elif tool_choice == 'none':
            assert last_message.content
            assert not last_message.tool_calls
        else:
            assert last_message.tool_calls
            called_tools = [tool_call.tool_name for tool_call in last_message.tool_calls]
            assert tool_choice in called_tools


        for message in new_messages:
            print(f'\n{message.role}:\n{message.content or message.tool_calls}\n')            
    print()


@base_test_all_providers
@pytest.mark.parametrize("messages, tools, tool_callables, tool_choice", [
    (
        [
            {"role": "system", "content": "Your role is use the available tools to answer questions like a cartoon Pirate"},            
            {"role": "user", "content": "What's the weather like in San Francisco? Respond in Fahrenheit."},
        ],
        [
            TOOLS["get_current_weather"], TOOLS["return_weather_messages"]
        ],
        {
            "get_current_weather": get_current_weather
        }, 
        ["get_current_weather", "return_weather_messages"]
    ),
])
def test_chat_enforce_tool_choice_sequence(
    # ai: UnifAIClient, 
    provider: AIProvider, 
    client_kwargs: dict,
    func_kwargs: dict,
    messages: list,
    tools: list,
    tool_callables: dict, 
    tool_choice: list[str]
    ):

    ai = UnifAIClient(
        {provider: client_kwargs},
        tool_callables=tool_callables
    )
    ai.init_client(provider, **client_kwargs)

    for _ in range(1):
        new_messages = ai.chat(
            messages=messages,
            provider=provider,
            tools=tools,
            tool_choice=tool_choice,
            return_on=tool_choice[-1],
            enforce_tool_choice=True,
            **func_kwargs
        )
        
        assert new_messages
        assert isinstance(new_messages, list)

        last_message = new_messages[-1]
        assert isinstance(last_message, Message)

        if tool_choice == 'auto':
            assert last_message.content or last_message.tool_calls
        elif tool_choice == 'required':
            assert last_message.tool_calls
            assert not last_message.content
        elif tool_choice == 'none':
            assert last_message.content
            assert not last_message.tool_calls
        else:
            assert last_message.tool_calls
            called_tools = [tool_call.tool_name for tool_call in last_message.tool_calls]
            assert tool_choice[-1] in called_tools


        for message in new_messages:
            print(f'\n{message.role}:\n{message.content or message.tool_calls}\n')            
    print()


from .baseaiclientwrapper import BaseAIClientWrapper
from .anthropic_wrapper import AnthropicWrapper
from .openai_wrapper import OpenAIWrapper
from .ollama_wrapper import OllamaWrapper
from .unifai import UnifAIClient, AIProvider


import pytest
from unifai import UnifAIClient, AIProvider
from unifai._types import (
    Message, 
    Image, 
    ToolCall, 
    ToolParameter,
    StringToolParameter,
    NumberToolParameter,
    IntegerToolParameter,
    BooleanToolParameter,
    NullToolParameter,
    ObjectToolParameter,
    ArrayToolParameter,
    Tool,
    FunctionTool,
    CodeInterpreterTool,
    FileSearchTool,
)

from basetest import base_test_all_providers

ai = UnifAIClient()

@pytest.mark.parametrize("input_messages, std_messages", [
    (
        [Message(role='user', content='Hello AI')],
        [Message(role='user', content='Hello AI')]
    ),
    (
        [{'role': 'user', 'content': 'Hello AI'}],
        [Message(role='user', content='Hello AI')]
    ),    
    (
        ['Hello AI'],
        [Message(role='user', content='Hello AI')]
    ),
    (
        [
            Message(role='system', content='Your role is...'), 
            Message(role='user', content='Hello AI'),
            Message(role='assistant', content='Hello User'),
        ],    
        [
            Message(role='system', content='Your role is...'), 
            Message(role='user', content='Hello AI'),
            Message(role='assistant', content='Hello User'),
        ],          
    ),
    (
        [
            {'role': 'system', 'content': 'Your role is...'}, 
            {'role': 'user', 'content': 'Hello AI'},
            {'role': 'assistant', 'content': 'Hello User'},
        ],
        [
            Message(role='system', content='Your role is...'), 
            Message(role='user', content='Hello AI'),
            Message(role='assistant', content='Hello User'),
        ],  
    ),    
    (
        [
            {'role': 'system', 'content': 'Your role is...'}, 
            'Hello AI',
            {'role': 'assistant', 'content': 'Hello User'},
        ],
        [
            Message(role='system', content='Your role is...'), 
            Message(role='user', content='Hello AI'),
            Message(role='assistant', content='Hello User'),
        ],  
    ),
    (
        [
            Message(role='system', content='Your role is...'),
            'Hello AI',
            {'role': 'assistant', 'content': 'Hello User'},
        ],
        [
            Message(role='system', content='Your role is...'), 
            Message(role='user', content='Hello AI'),
            Message(role='assistant', content='Hello User'),
        ],  
    ),
])
def test_standardize_messages(input_messages, std_messages):
    assert ai.standardize_messages(input_messages) == std_messages



TOOL_DICTS = {
    "code_interpreter": {"type": "code_interpreter"},
    "file_search": {"type": "file_search"},
    "get_current_weather": {
        "type": "function",
        "function": {
            "name": "get_current_weather",
            "description": "Get the current weather in a given location",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The city and state, e.g. San Francisco, CA",
                    },
                    "unit": {
                        "type": "string", 
                        "enum": ["celsius", "fahrenheit"]
                    }
                },
                "required": ["location"],
                "additionalProperties": False
            },
            "strict": True
        },
    }, # get_current_weather
    "get_object_with_all_types": {
        "type": "function",
        "function": {
            "name": "get_object_with_all_types",
            "description": "Get an object with all types",
            "parameters": {
                "type": "object",
                "properties": {
                    "string_param": {
                        "type": "string",
                        "description": "A string parameter",
                        "enum": ["a", "b", "c"]
                    },
                    "number_param": {
                        "type": "number",
                        "description": "A number parameter",
                        "enum": [1.0, 2.0, 3.0]
                    },
                    "integer_param": {
                        "type": "integer",
                        "description": "An integer parameter",
                        "enum": [1, 2, 3]
                    },
                    "boolean_param": {
                        "type": "boolean",
                        "description": "A boolean parameter",
                        # "enum": [True, False]
                    },
                    "null_param": {
                        "type": "null",
                        "description": "A null parameter",
                        # "enum": [None]
                    },
                    "array_param": {
                        "type": "array",
                        "description": "An array parameter",
                        "items": {
                            "type": "string",
                            "description": "A string item",
                            "enum": ["a", "b", "c"]                            
                        }
                    },
                    "object_param": {
                        "type": "object",
                        "description": "An object parameter",
                        "properties": {
                            "string_prop": {
                                "type": "string",
                                "description": "A string property",
                                "enum": ["a", "b", "c"]
                            },
                            "number_prop": {
                                "type": "number",
                                "description": "A number property",
                                "enum": [1.0, 2.0, 3.0]
                            },
                            "integer_prop": {
                                "type": "integer",
                                "description": "An integer property",
                                "enum": [1, 2, 3]
                            },
                            "boolean_prop": {
                                "type": "boolean",
                                "description": "A boolean property",
                                # "enum": [True, False]
                            },
                            "null_prop": {
                                "type": "null",
                                "description": "A null property",
                                # "enum": [None]
                            },
                            "array_prop": {
                                "type": "array",
                                "description": "An array property",
                                "items": {
                                    "type": "string",
                                    "description": "A string item",
                                    "enum": ["a", "b", "c"]                            
                                }
                            },
                            "nested_object_prop": {
                                "type": "object",
                                "description": "A nested object property",
                                "properties": {
                                    "nested_string_prop": {
                                        "type": "string",
                                        "description": "A string property in a nested object",
                                        "enum": ["a", "b", "c"]
                                    },
                                    "nested_number_prop": {
                                        "type": "number",
                                        "description": "A number property in a nested object",
                                        "enum": [1.0, 2.0, 3.0]
                                    },
                                    "nested_integer_prop": {
                                        "type": "integer",
                                        "description": "An integer property in a nested object",
                                        "enum": [1, 2, 3]
                                    },
                                    "nested_boolean_prop": {
                                        "type": "boolean",
                                        "description": "A boolean property in a nested object",
                                        # "enum": [True, False]
                                    },
                                    "nested_null_prop": {
                                        "type": "null",
                                        "description": "A null property in a nested object",
                                        # "enum": [None]
                                    },
                                    "nested_array_prop": {
                                        "type": "array",
                                        "description": "An array property in a nested object",
                                        "items": {
                                            "type": "string",
                                            "description": "A string item in array in a nested object",
                                            "enum": ["a", "b", "c"]                            
                                        }
                                    },
                                },
                                "required": ["nested_string_prop", "nested_number_prop", "nested_integer_prop", "nested_boolean_prop", "nested_null_prop", "nested_array_prop"],
                                "additionalProperties": False
                            },
                        },
                        "required": ["string_prop", "number_prop", "integer_prop", "boolean_prop", "null_prop", "array_prop", "nested_object_prop"],
                        "additionalProperties": False
                    }
                },
                "required": ["string_param", "number_param", "integer_param", "boolean_param", "null_param", "array_param", "object_param"],
                "additionalProperties": False
            },
            "strict": True
        }
    }, # get_object_with_all_types
}    

TOOL_OBJECTS = {
    "code_interpreter": CodeInterpreterTool(),
    "file_search": FileSearchTool(),    
    "get_current_weather": FunctionTool(
        name="get_current_weather",
        description="Get the current weather in a given location",
        parameters=ObjectToolParameter(
            # name="parameters",
            properties=[
                StringToolParameter(
                    name="location",
                    description="The city and state, e.g. San Francisco, CA",
                    required=True
                ),
                StringToolParameter(
                    name="unit",
                    enum=["celsius", "fahrenheit"]
                )
            ],
        )
    ),
    "get_object_with_all_types": FunctionTool(
        name="get_object_with_all_types",
        description="Get an object with all types",
        parameters=ObjectToolParameter(
            properties=[
                StringToolParameter(
                    name="string_param",
                    description="A string parameter",
                    required=True,
                    enum=["a", "b", "c"]
                ),
                NumberToolParameter(
                    name="number_param",
                    description="A number parameter",
                    required=True,
                    enum=[1.0, 2.0, 3.0]
                ),
                IntegerToolParameter(
                    name="integer_param",
                    description="An integer parameter",
                    required=True,
                    enum=[1, 2, 3]
                ),
                BooleanToolParameter(
                    name="boolean_param",
                    description="A boolean parameter",
                    required=True,
                ),
                NullToolParameter(
                    name="null_param",
                    description="A null parameter",
                    required=True,
                ),
                ArrayToolParameter(
                    name="array_param",
                    description="An array parameter",
                    required=True,
                    items=StringToolParameter(
                        description="A string item",
                        required=True,
                        enum=["a", "b", "c"]
                    )
                ),
                ObjectToolParameter(
                    name="object_param",
                    description="An object parameter",
                    required=True,
                    properties=[
                        StringToolParameter(
                            name="string_prop",
                            description="A string property",
                            required=True,
                            enum=["a", "b", "c"]
                        ),
                        NumberToolParameter(
                            name="number_prop",
                            description="A number property",
                            required=True,
                            enum=[1.0, 2.0, 3.0]
                        ),
                        IntegerToolParameter(
                            name="integer_prop",
                            description="An integer property",
                            required=True,
                            enum=[1, 2, 3]
                        ),
                        BooleanToolParameter(
                            name="boolean_prop",
                            description="A boolean property",
                            required=True,
                        ),
                        NullToolParameter(
                            name="null_prop",
                            description="A null property",
                            required=True,
                        ),
                        ArrayToolParameter(
                            name="array_prop",
                            description="An array property",
                            required=True,
                            items=StringToolParameter(
                                description="A string item",
                                required=True,
                                enum=["a", "b", "c"]
                            )
                        ),
                        ObjectToolParameter(
                            name="nested_object_prop",
                            description="A nested object property",
                            required=True,
                            properties=[
                                StringToolParameter(
                                    name="nested_string_prop",
                                    description="A string property in a nested object",
                                    required=True,
                                    enum=["a", "b", "c"]
                                ),
                                NumberToolParameter(
                                    name="nested_number_prop",
                                    description="A number property in a nested object",
                                    required=True,
                                    enum=[1.0, 2.0, 3.0]
                                ),
                                IntegerToolParameter(
                                    name="nested_integer_prop",
                                    description="An integer property in a nested object",
                                    required=True,
                                    enum=[1, 2, 3]
                                ),
                                BooleanToolParameter(
                                    name="nested_boolean_prop",
                                    description="A boolean property in a nested object",
                                    required=True,
                                ),
                                NullToolParameter(
                                    name="nested_null_prop",
                                    description="A null property in a nested object",
                                    required=True,
                                ),
                                ArrayToolParameter(
                                    name="nested_array_prop",
                                    description="An array property in a nested object",
                                    required=True,
                                    items=StringToolParameter(
                                        description="A string item in array in a nested object",
                                        required=True,
                                        enum=["a", "b", "c"]
                                    )
                                ),
                            ],
                        ), # nested_object_prop
                    ],
                ) # object_param
            ],
        )
    )
            
}

@pytest.mark.parametrize("input_tools, expected_std_tools", [
    (
        [TOOL_DICTS["code_interpreter"]],
        [TOOL_OBJECTS["code_interpreter"]]
    ), 
    (
        [TOOL_DICTS["file_search"]],
        [TOOL_OBJECTS["file_search"]],
    ),     
    (
        [TOOL_DICTS["get_current_weather"]],
        [TOOL_OBJECTS["get_current_weather"]]
    ),    
    (
        [TOOL_DICTS["get_object_with_all_types"]],
        [TOOL_OBJECTS["get_object_with_all_types"]]
    ),
    (
        # test all tools
        list(TOOL_DICTS.values()),
        list(TOOL_OBJECTS.values())
    )
    
])
def test_standardize_tools(input_tools, expected_std_tools):
    std_tools = ai.standardize_tools(input_tools) 
    dict_tools = [tool.to_dict() for tool in std_tools]

    # assert len(std_tools) == len(expected_std_tools)
    # for std_tool, expected_tool in zip(std_tools, expected_std_tools):
    #     for std_param, expected_param in zip(std_tool.parameters.properties, expected_tool.parameters.properties):
    #         assert std_param == expected_param

    #     assert std_tool.name == expected_tool.name
    #     assert std_tool.description == expected_tool.description
    #     assert std_tool.parameters.required == expected_tool.parameters.required
    #     assert std_tool == expected_tool

    assert std_tools == expected_std_tools
    assert dict_tools == input_tools

    


from typing import Type, Optional, Sequence, Any, Union, Literal, TypeVar

from json import dumps as json_dumps
from ._types import Message, Tool, ToolCall, Image

T = TypeVar("T")

class BaseAIClientWrapper:
    default_model = "mistral:7b-instruct"

    def import_client(self) -> Type:
        raise NotImplementedError("This method must be implemented by the subclass")
    
    def init_client(self, **client_kwargs) -> Any:
        self._client = self.import_client()(**client_kwargs)
        return self._client

    def __init__(self, **client_kwargs):
        self._client = None
        self.client_kwargs = client_kwargs

    @property
    def client(self) -> Type:
        if self._client is None:
            return self.init_client(**self.client_kwargs)
        return self._client
    


    def prep_input_message(self, message: Message) -> Any:
        raise NotImplementedError("This method must be implemented by the subclass")
    
    def prep_input_messages_and_system_prompt(self, 
                                              messages: list[Message], 
                                              system_prompt_arg: Optional[str] = None
                                              ) -> tuple[list, Optional[str]]:
        raise NotImplementedError("This method must be implemented by the subclass")    
    
    def prep_input_tool(self, tool: Tool) -> Any:
        raise NotImplementedError("This method must be implemented by the subclass")
        
    def prep_input_tool_call(self, tool_call: ToolCall) -> Any:
        raise NotImplementedError("This method must be implemented by the subclass")

    def prep_input_tool_call_response(self, tool_call: ToolCall, tool_response: Any) -> Any:
        raise NotImplementedError("This method must be implemented by the subclass")
    
    def split_tool_call_outputs_into_messages(self, tool_calls: list[ToolCall]) -> list[Message]:
        raise NotImplementedError("This method must be implemented by the subclass")


    def prep_input_message_tool_call(self, message: Message, input_object: T) -> T:
        raise NotImplementedError("This method must be implemented by the subclass")

    def prep_input_message_tool_call_response(self, message: Message, input_object: T) -> T:
        raise NotImplementedError("This method must be implemented by the subclass")
        
    
    def prep_input_tool_choice(self, tool_choice: str) -> Any:
        raise NotImplementedError("This method must be implemented by the subclass")    

    def prep_input_response_format(self, response_format: Union[str, dict]) -> Any:
        raise NotImplementedError("This method must be implemented by the subclass")

    def extract_output_tool_call(self, tool_call: Any) -> ToolCall:
        raise NotImplementedError("This method must be implemented by the subclass")
    
    def extract_std_and_client_messages(self, response: Any) -> tuple[Message, Any]:
        raise NotImplementedError("This method must be implemented by the subclass")   

    # List Models
    def list_models(self) -> list[str]:
        raise NotImplementedError("This method must be implemented by the subclass")    


    # Chat 
    def chat(
            self,
            messages: list[Message],     
            model: Optional[str] = None,
            system_prompt: Optional[str] = None,                   
            tools: Optional[list[Any]] = None,
            tool_choice: Optional[Union[Tool, str, dict, Literal["auto", "required", "none"]]] = None,            
            response_format: Optional[Union[str, dict[str, str]]] = None,
            **kwargs
            ) -> Message:
        raise NotImplementedError("This method must be implemented by the subclass")

    # Embeddings

    def embeddings(
            self,
            model: Optional[str] = None,
            texts: Optional[Sequence[str]] = None,
            **kwargs
            ):
        raise NotImplementedError("This method must be implemented by the subclass")
    
    # Generate

    def generate(
            self,
            model: Optional[str] = None,
            prompt: Optional[str] = None,
            **kwargs
            ):
        raise NotImplementedError("This method must be implemented by the subclass")


    def create_assistant(self, **kwargs):
        raise NotImplementedError("This method must be implemented by the subclass")
        
    def update_assistant(self, ass_id, **kwargs):
        raise NotImplementedError("This method must be implemented by the subclass")
    

    def create_thread(self):
        raise NotImplementedError("This method must be implemented by the subclass")
    
    def create_run(self):
        raise NotImplementedError("This method must be implemented by the subclass")    


import pytest
from unifai import UnifAIClient, AIProvider
from unifai._types import Message, Tool, EvalTypeParameters
from basetest import base_test_all_providers

TOOLS = {
    "return_flagged_and_reason": {
        "type": "function",
                "function": {
                    "name": "return_flagged_and_reason",
                    "description": "Return a boolean indicating whether the content should be flagged and a concise reason for the flag if True.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "flagged": {
                                "type": "boolean",
                                "description": "True if the content should be flagged, False otherwise."
                            },
                            "reason": {
                                "type": "string",
                                "description": "A concise reason for the flag if True. An empty string if False."
                            }
                        },
                        "required": ["flagged", "reason"]
                    }
                }
    },
}

EVAL_TYPES = [
    EvalTypeParameters(
        name="urlEval",
        system_prompt=(
        "You review HTML text to flag elements that may contain spam, misinformation, or other malicious items. "
        "You also check the associated URLS for signs of typosquatting or spoofing. "
        # "Use the return_flagged_and_reason function to return your result."
        ),
        tools=["return_flagged_and_reason"],
        tool_choice="return_flagged_and_reason",
    )
]

@base_test_all_providers
@pytest.mark.parametrize("tools, tool_callables, eval_types, eval_type, content", [
    ([TOOLS["return_flagged_and_reason"]], None, EVAL_TYPES, "urlEval", {"url": "https://google.com", "link_text": "Google"}),
    ([TOOLS["return_flagged_and_reason"]], None, EVAL_TYPES, "urlEval", {"url": "https://g00gle.com", "link_text": "Google"}),
])
def test_evaluate_simple(
    provider: AIProvider, 
    client_kwargs: dict, 
    func_kwargs: dict,
    tools: list,
    tool_callables: dict,
    eval_types: list,
    eval_type: str,
    content: str
    ):

    ai = UnifAIClient(
        provider_client_kwargs={provider: client_kwargs},
        tools=tools,
        tool_callables=tool_callables,
        eval_types=eval_types
    )
    ai.init_client(provider, **client_kwargs)

    response = ai.evaluate(
        eval_type=eval_type, 
        content=content,
        provider=provider,
        **func_kwargs
        )
    print(response)
    assert response



import pytest
from unifai import UnifAIClient
from basetest import base_test_all_providers

@base_test_all_providers
def test_list_models(provider, client_kwargs, func_kwargs):
    ai = UnifAIClient({provider: client_kwargs})
    for provider_arg in [provider, None]:
        models = ai.list_models(provider_arg)
        assert models
        assert isinstance(models, list)
        assert isinstance(models[0], str)
        print(f'{provider} Models: \n' + '\n'.join(models))




import pytest
from unifai import UnifAIClient

def test_import_simplifai_client():
    assert UnifAIClient

@pytest.mark.parametrize("provider, client_kwargs", [
    ("anthropic", {"api_key": "test"}),
    ("openai", {"api_key": "test"}),
    ("ollama", {}),
])
def test_init_clients(provider, client_kwargs):
    ai = UnifAIClient({
        provider: client_kwargs
    })

    assert ai.provider_client_kwargs[provider] == client_kwargs
    assert ai.providers == [provider]
    assert ai._clients == {}
    assert ai.default_provider == provider

    wrapper_name = provider.capitalize().replace("ai", "AI") + "Wrapper"
    # assert wrapper_name not in globals()

    client = ai.init_client(provider)    
    # assert wrapper_name in globals()
    # wrapper = globals()[wrapper_name]

    assert client
    # assert isinstance(client, wrapper)    
    assert ai._clients[provider] == client
    assert ai.get_client(provider) == client
    assert ai.get_client() == client
    


from typing import Optional, Union, Sequence, Any, Literal, Mapping

from anthropic import Anthropic
from anthropic.types import (
    Usage as AnthropicUsage,    
    Message as AnthropicMessage,
    MessageParam as AnthropicMessageParam,
    ContentBlock as AnthropicContentBlock,
    TextBlock as AnthropicTextBlock,
    ToolUseBlock as AnthropicToolUseBlock,
    TextBlockParam as AnthropicTextBlockParam,
    ImageBlockParam as AnthropicImageBlockParam,
    ToolUseBlockParam as AnthropicToolUseBlockParam,
    ToolResultBlockParam as AnthropicToolResultBlockParam,
    ToolParam as AnthropicToolParam, 
)
from anthropic.types.image_block_param import Source as AnthropicImageSource



from ._types import Message, Tool, ToolCall, Image
from ._convert_types import stringify_content
from .baseaiclientwrapper import BaseAIClientWrapper
from io import BytesIO

class AnthropicWrapper(BaseAIClientWrapper):
    client: Anthropic
    default_model = "claude-3-5-sonnet-20240620"

    def import_client(self):
        from anthropic import Anthropic
        return Anthropic


    def prep_input_message(self, message: Message) -> AnthropicMessageParam:
        if message.role == "system":
            raise ValueError("Anthropic does not support system messages")
        
        role = "user" if message.role != "assistant" else message.role
        content = []


        if message.tool_calls:
            if message.role == "tool":
                content = self.prep_input_message_tool_call_response(message, content)
            else:
                content = self.prep_input_message_tool_call(message, content)

        if message.content and not content:
            content.append(AnthropicTextBlockParam(text=message.content, type="text"))            

        if message.images and not message.tool_calls:
            # TODO: Implement image handling
            for image in message.images:
                source = AnthropicImageSource(data=image.data, media_type=image.media_type, type=image.format), 
                content.append(AnthropicImageBlockParam(source=source, type="image"))

        return AnthropicMessageParam(role=role, content=content)     
            

    def prep_input_messages_and_system_prompt(self, 
                                              messages: list[Message], 
                                              system_prompt_arg: Optional[str] = None
                                              ) -> tuple[list, Optional[str]]:
        
        system_prompt = system_prompt_arg
        if messages and messages[0].role == "system":
            # Remove the first system message from the list since Anthropic does not support system messages
            system_message = messages.pop(0)
            # Set the system prompt to the content of the first system message if not set by the argument
            if not system_prompt:
                system_prompt = system_message.content 

        client_messages = [self.prep_input_message(message) for message in messages]
        return client_messages, system_prompt            
    
    def prep_input_tool(self, tool: Tool) -> AnthropicToolParam:
        return AnthropicToolParam(
            name=tool.name,
            description=tool.description,
            input_schema=tool.parameters.to_dict()
        )
        
    # def prep_input_tool_call(self, tool_call: ToolCall) -> AnthropicToolUseBlockParam:
    #     return AnthropicToolUseBlockParam(
    #         id=tool_call.tool_call_id,
    #         name=tool_call.tool_name,
    #         input=tool_call.arguments,
    #         type="tool_use"
    #     )

    # def prep_input_tool_call_response(self, tool_call: ToolCall, tool_response: Any) -> AnthropicToolResultBlockParam:
    #     raise NotImplementedError("This method must be implemented by the subclass")


    def prep_input_message_tool_call(self, message: Message, input_object: list) -> list:
        if message.tool_calls and message.role == "assistant":
            for tool_call in message.tool_calls:
                tool_use_param = AnthropicToolUseBlockParam(
                    id=tool_call.tool_call_id,
                    name=tool_call.tool_name,
                    input=tool_call.arguments,
                    type="tool_use"
                )
                input_object.append(tool_use_param)                
        return input_object

    
    def split_tool_call_outputs_into_messages(self, tool_calls: list[ToolCall]) -> list[Message]:        
        message = Message(
            role="tool",
            tool_calls=tool_calls,            
        )
        return [message]
    
    def prep_input_message_tool_call_response(self, message: Message, input_object: list) -> list:
        if message.tool_calls and message.role == "tool":
            for tool_call in message.tool_calls:
                content = []
                if message.content:
                    content.append(AnthropicTextBlockParam(text=message.content, type="text"))
                if message.images:
                    for image in message.images:
                        source = AnthropicImageSource(data=image.data, media_type=image.media_type, type=image.format), 
                        content.append(AnthropicImageBlockParam(source=source, type="image"))
                
                tool_result_param = AnthropicToolResultBlockParam(
                    tool_use_id=tool_call.tool_call_id,
                    content=content,
                    is_error=False,
                    type="tool_result"
                )
                input_object.append(tool_result_param)
        return input_object


    def prep_input_tool_choice(self, tool_choice: str) -> dict:
        if tool_choice == "any":
            tool_choice = "required"
        if tool_choice in ("auto", "required", "none"):
            return {"type": tool_choice}
        
        return {"type": "tool", "name": tool_choice}

    def prep_input_response_format(self, response_format: Union[str, dict]) -> None:
        # Warn: response_format is not used by the Anthropic client
        if response_format: print("Warning: response_format is not used by the Anthropic client")
        return None

            
    def extract_output_tool_call(self, tool_call: AnthropicToolUseBlock) -> ToolCall:
        return ToolCall(
            tool_call_id=tool_call.id,
            tool_name=tool_call.name,
            arguments=tool_call.input
        )
    
    def extract_std_and_client_messages(self, response: AnthropicMessage) -> tuple[Message, AnthropicMessageParam]:
        client_message = AnthropicMessageParam(role=response.role, content=response.content)
        content = ""
        tool_calls = []
        for block in response.content:
            if block.type == "text":
                content += block.text
            elif block.type == "tool_use":
                tool_calls.append(self.extract_output_tool_call(block))

        std_message = Message(
            role=response.role,
            content=content,
            tool_calls=tool_calls,
            response_object=response
        )

        return std_message, client_message

        
    # List Models
    def list_models(self) -> list[str]:
        claude_models = [
            "claude-3-5-sonnet-20240620",
            "claude-3-opus-20240229",
            "claude-3-sonnet-20240229",
            "claude-3-haiku-20240307",
            "claude-2.1",
            "claude-2.0",
            "claude-instant-1.2",
        ]
        return claude_models
        


    # Chat 
    def chat(
            self,
            messages: list[AnthropicMessageParam],     
            model: str = default_model, 
            system_prompt: Optional[str] = None,                 
            tools: Optional[list[AnthropicToolParam]] = None,
            tool_choice: Optional[dict] = None,
            response_format: Optional[Union[str, dict[str, str]]] = None,
            **kwargs
            ) -> AnthropicMessage:
        
        max_tokens = kwargs.pop("max_tokens", 4096)
        if system_prompt:
            kwargs["system"] = system_prompt
        if tools:
            kwargs["tools"] = tools
            if tool_choice:
                kwargs["tool_choice"] = tool_choice

        response = self.client.messages.create(
            max_tokens=max_tokens,
            messages=messages,
            model=model,
            **kwargs
        )
        return response

    # Embeddings

    def embeddings(
            self,
            model: Optional[str] = None,
            texts: Optional[Sequence[str]] = None,
            **kwargs
            ):
        raise NotImplementedError("This method must be implemented by the subclass")
    
    # Generate

    def generate(
            self,
            model: Optional[str] = None,
            prompt: Optional[str] = None,
            **kwargs
            ):
        raise NotImplementedError("This method must be implemented by the subclass")


    def create_assistant(self, **kwargs):
        raise NotImplementedError("This method must be implemented by the subclass")
        
    def update_assistant(self, ass_id, **kwargs):
        raise NotImplementedError("This method must be implemented by the subclass")
    

    def create_thread(self):
        raise NotImplementedError("This method must be implemented by the subclass")
    
    def create_run(self):
        raise NotImplementedError("This method must be implemented by the subclass")    


from typing import Optional, Union, Sequence, Any, Literal, Mapping

from ._types import Message, Tool, ToolCall
from ._convert_types import stringify_content
from .baseaiclientwrapper import BaseAIClientWrapper

from ollama import Client as OllamaClient
from ollama._types import (
    ChatResponse as OllamaChatResponse,
    Message as OllamaMessage,
    Tool as OllamaTool, 
    ToolFunction as OllamaToolFunction,
    ToolCallFunction as OllamaToolCallFunction,
    ToolCall as OllamaToolCall, 
    Parameters as OllamaParameters, 
    Property as OllamaProperty,
    Options as OllamaOptions
)


from random import choices as random_choices
from string import ascii_letters, digits
from hashlib import sha256

def generate_random_id(length=8):
    return ''.join(random_choices(ascii_letters + digits, k=length))

def sha256_hash(string: str) -> str:
    return sha256(string.encode()).hexdigest()

class OllamaWrapper(BaseAIClientWrapper):
    client: OllamaClient

    def import_client(self):
        from ollama import Client as OllamaClient
        from ollama._types import Tool, ToolFunction, Parameters, Property, Message

        return OllamaClient
    
    def prep_input_message(self, message: Message) -> OllamaMessage:        
        role = message.role
        content = message.content or ''
        images = None # TODO: Implement image extraction
        tool_calls = None
        if message.images:
            images = [image.data for image in message.images if image.data]
        if message.tool_calls:
            if message.role == 'tool':
                tool_calls = None
            else:
                tool_calls = [self.prep_input_tool_call(tool_call) for tool_call in message.tool_calls]
            
        return OllamaMessage(role=role, content=content, images=images, tool_calls=tool_calls)

    def prep_input_messages_and_system_prompt(self, 
                                              messages: list[Message], 
                                              system_prompt_arg: Optional[str] = None
                                              ) -> tuple[list, Optional[str]]:
        if system_prompt_arg:
            system_prompt = system_prompt_arg
            if messages and messages[0].role == "system":
                messages[0].content = system_prompt
            else:
                messages.insert(0, Message(role="system", content=system_prompt))
        elif messages and messages[0].role == "system":
            system_prompt = messages[0].content
        else:
            system_prompt = None

        client_messages = [self.prep_input_message(message) for message in messages]
        return client_messages, system_prompt
            
    
    def prep_input_tool(self, tool: Tool) -> OllamaTool:
        # tool_dict = tool if isinstance(tool, dict) else tool.to_dict()

        tool_dict = tool.to_dict()
        tool_type = tool_dict['type']
        tool_def = tool_dict[tool_type]

        tool_name = tool_def['name']
        tool_description = tool_def['description']
        tool_parameters = tool_def['parameters']
        parameters_type = tool_parameters['type']
        parameters_required = tool_parameters['required']

        properties = {}
        for prop_name, prop_def in tool_parameters['properties'].items():
            prop_type = prop_def['type']
            prop_description = prop_def.get('description', None)
            prop_enum = prop_def.get('enum', None)
            properties[prop_name] = OllamaProperty(type=prop_type, description=prop_description, enum=prop_enum)
        
        parameters = OllamaParameters(type=parameters_type, required=parameters_required, properties=properties)
        tool_function = OllamaToolFunction(name=tool_name, description=tool_description, parameters=parameters)
        return OllamaTool(type=tool_type, function=tool_function)            
    
    def prep_input_tool_choice(self, tool_choice: str) -> str:
        return tool_choice
    
    def prep_input_tool_call(self, tool_call: ToolCall) -> OllamaToolCall:
        return OllamaToolCall(
            function=OllamaToolCallFunction(
                name=tool_call.tool_name, 
                arguments=tool_call.arguments if tool_call.arguments else {}
            )
        )
    
    
    def split_tool_call_outputs_into_messages(self, tool_calls: list[ToolCall]) -> list[Message]:        
        return [
            Message(role="tool", content=stringify_content(tool_call.output), tool_calls=[tool_call])
            for tool_call in tool_calls
        ]

    # def prep_input_tool_call_response(self, tool_call: ToolCall, tool_response: Any) -> OllamaMessage:
    #     return OllamaMessage(
    #         role='tool',
    #         content=tool_response,
    #     )

    def prep_input_response_format(self, response_format: Union[str, dict[str, str]]) -> str:
        if response_format in ("json", "json_object"):
            return 'json'
        return ''

    def extract_output_tool_call(self, tool_call: OllamaToolCall) -> ToolCall:
        tool_call_function = tool_call['function']
        name = tool_call_function['name']
        arguments = tool_call_function.get('arguments')
        tool_call_id = f'call_{generate_random_id(24)}'
        return ToolCall(tool_call_id=tool_call_id, tool_name=name, arguments=arguments)


    # def extract_output_message(self, response: OllamaChatResponse) -> Message:
    #     message = response['message']
    #     images = None # TODO: Implement image extraction
    #     tool_calls = [self.extract_output_tool_call(tool_call) for tool_call in message.get('tool_calls', [])]
    #     return Message(
    #         role=message['role'],
    #         content=message.get('content'),
    #         images=images,
    #         tool_calls=tool_calls,
    #         response_object=response
    #     )

    def extract_std_and_client_messages(self, response: OllamaChatResponse) -> tuple[Message, OllamaMessage]:
        client_message = response['message']
        images = None # TODO: Implement image extraction
        tool_calls = [self.extract_output_tool_call(tool_call) for tool_call in client_message.get('tool_calls', [])]
        std_message = Message(
            role=client_message['role'],
            content=client_message.get('content'),
            images=images,
            tool_calls=tool_calls,
            response_object=response
        )
        return std_message, client_message




    def list_models(self) -> Sequence[str]:
        return [model_name for model_dict in self.client.list()["models"] if (model_name := model_dict.get("name"))]
        
    def get_custom_model(self, base_model: str, system_prompt: str) -> str:
        model_name = f"{base_model}_{sha256_hash(system_prompt)}"
        if model_name not in self.list_models():
            modelfile = f'FROM {base_model}\nSYSTEM """{system_prompt}"""'
            prog_response = self.client.create(model=model_name, modelfile=modelfile)
            print(prog_response)
        return model_name
    
    def get_system_prompt_from_messages(self, messages: list[Message]) -> Optional[str]:
        if messages and messages[0].role == 'system':
            return messages[0].content


    def chat(
            self,
            messages: list[OllamaMessage],     
            model: Optional[str] = None, 
            system_prompt: Optional[str] = None,                   
            tools: Optional[list[OllamaTool]] = None,
            tool_choice: Optional[str] = None,
            response_format: Optional[str] = '',
            **kwargs
            ) -> Message:
        
            # if (tool_choice and tool_choice != "auto" and system_prompt
            #     and messages and messages[0]["role"] == "system"
            #     ):                
            #     if tool_choice == "required":
            #         messages[0]["content"] = f"{system_prompt}\nYou MUST call one or more tools."
            #     elif tool_choice == "none":
            #         messages[0]["content"] = f"{system_prompt}\nYou CANNOT call any tools."
            #     else:
            #         messages[0]["content"] = f"{system_prompt}\nYou MUST call the tool '{tool_choice}' with ALL of its required arguments."
            #     system_prompt_modified = True
            # else:
            #     system_prompt_modified = False

            user_messages = [message for message in messages if message["role"] == 'user']
            last_user_content = user_messages[-1].get("content", "") if user_messages else ""            
            if (tool_choice and tool_choice != "auto" and last_user_content is not None
                ):                
                if tool_choice == "required":
                    user_messages[-1]["content"] = f"{last_user_content}\nYou MUST call one or more tools."
                elif tool_choice == "none":
                    user_messages[-1]["content"] = f"{last_user_content}\nYou CANNOT call any tools."
                else:
                    user_messages[-1]["content"] = f"{last_user_content}\nYou MUST call the tool '{tool_choice}' with ALL of its required arguments."
                last_user_content_modified = True
            else:
                last_user_content_modified = False

            keep_alive = kwargs.pop('keep_alive', None)
            stream = kwargs.pop('stream', False)
            options = kwargs.pop('options', None)
            if not options and kwargs:
                options = OllamaOptions(**kwargs)

            response = self.client.chat(
                messages=messages,                 
                model=model, 
                tools=tools, 
                format=response_format, 
                keep_alive=keep_alive,
                stream=stream,
                options=options
            )

            if last_user_content_modified:
                user_messages[-1]["content"] = last_user_content

            # if system_prompt_modified:
            #     response['message']['content'] = system_prompt

            return response


from json import dumps as json_dumps
from typing import Optional, Union, Sequence, Any, Literal, Mapping, TypeVar, List, Type, Tuple, Dict, Any
from ._types import (
    Message,
    Tool,
    FunctionTool,
    CodeInterpreterTool,
    FileSearchTool,
    ToolCall,
    ToolParameter, 
    ToolDict, 
    ToolParameterType, 
    ToolValPyTypes, 
    StringToolParameter, 
    NumberToolParameter, 
    IntegerToolParameter,
    BooleanToolParameter,
    NullToolParameter,
    ArrayToolParameter,
    ObjectToolParameter,
    AnyOfToolParameter,
)

def make_content_serializeable(content: Any) -> Union[str, int, float, bool, dict, list, None]:
    """Recursively makes an object serializeable by converting it to a dict or list of dicts and converting all non-string values to strings."""
    if content is None or isinstance(content, (str, int, float, bool)):
        return content
    if isinstance(content, Mapping):
        return {k: make_content_serializeable(v) for k, v in content.items()}
    if isinstance(content, Sequence):
        return [make_content_serializeable(item) for item in content]
    return str(content) 

def stringify_content(content: Any) -> str:
    """Formats content for use a message content. If content is not a string, it is converted to a json string."""
    if isinstance(content, str):
        return content
    return json_dumps(make_content_serializeable(content), indent=0)

def make_few_shot_prompt(        
        system_prompt: Optional[str] = None, 
        examples: Optional[list[Union[Message, dict[Literal["input", "response"], Any]]]] = None, 
        content: Any = ""
    ) -> Sequence[Message]:
    """Makes list of message objects from system prompt, examples, and user input."""
    messages = []
    if system_prompt:
        # Begin with system_prompt if it exists
        messages.append(Message(role="system", content=system_prompt))

    # Add examples
    if examples:
        for example in examples:
            if isinstance(example, Message):
                messages.append(example)
            else:
                messages.append(Message(role="user", content=stringify_content(example['input'])))
                messages.append(Message(role="assistant", content=stringify_content(example['response'])))

    # Add content
    messages.append(Message(role="user", content=stringify_content(content)))
    return messages


def tool_parameter_from_dict(
        param_dict: dict, 
        param_name: Optional[str]= None,
        param_required: bool= False
        ) -> ToolParameter:
    
    if (anyof_param_dicts := param_dict.get('anyOf')) is not None:
        anyOf = [
            tool_parameter_from_dict(param_dict=anyof_param_dict, param_name=param_name, param_required=param_required)
            for anyof_param_dict in anyof_param_dicts
        ]
        return AnyOfToolParameter(name=param_name, required=param_required, anyOf=anyOf)

    param_type = param_dict['type']
    param_name = param_dict.get('name', param_name)
    param_description = param_dict.get('description')
    param_enum = param_dict.get('enum')

    if param_type == 'string':
        return StringToolParameter(name=param_name, description=param_description, required=param_required, enum=param_enum)
    if param_type == 'number':
        return NumberToolParameter(name=param_name, description=param_description, required=param_required, enum=param_enum)
    if param_type == 'integer':
        return IntegerToolParameter(name=param_name, description=param_description, required=param_required, enum=param_enum)
    if param_type == 'boolean':
        return BooleanToolParameter(name=param_name, description=param_description, required=param_required, enum=param_enum)
    if param_type == 'null':
        return NullToolParameter(name=param_name, description=param_description, required=param_required, enum=param_enum)
    
    if param_type == 'array':
        items = tool_parameter_from_dict(param_dict=param_dict['items'], param_required=param_required)
        return ArrayToolParameter(name=param_name, description=param_description, 
                                  required=param_required, enum=param_enum, 
                                  items=items)    
    if param_type == 'object':
        required_params = param_dict.get('required', [])
        properties = [
            tool_parameter_from_dict(param_dict=prop_dict, param_name=prop_name, param_required=prop_name in required_params) 
            for prop_name, prop_dict in param_dict['properties'].items()
        ]
        additionalProperties = param_dict.get('additionalProperties', False)
        return ObjectToolParameter(name=param_name, description=param_description, 
                                   required=param_required, enum=param_enum, 
                                   properties=properties, additionalProperties=additionalProperties)
    
    raise ValueError(f"Invalid parameter type: {param_type}")



def tool_from_dict(tool_dict: dict) -> Tool:
    tool_type = tool_dict['type']
    if tool_type == 'code_interpreter':
        return CodeInterpreterTool()
    if tool_type == 'file_search':
        return FileSearchTool()
    
    tool_def = tool_dict[tool_type]
    parameters = tool_parameter_from_dict(param_dict=tool_def['parameters'], 
            # param_name='parameters',
            # param_required=True
    )
    if parameters.type == 'anyOf':
        raise ValueError("Root parameter cannot be anyOf: See: https://platform.openai.com/docs/guides/structured-outputs/root-objects-must-not-be-anyof")

    return FunctionTool(
        name=tool_def['name'], 
        description=tool_def['description'], 
        parameters=parameters,
        strict=tool_def.get('strict', True)
    )


from typing import (
    Optional, Union, Sequence, Any, 
    Literal, Mapping, TypeVar, List, Type, Tuple, Dict, Any, Callable, Collection)

from pydantic import BaseModel, Field

class ResponseInfo(BaseModel):
    data: Any
    usage: Any

class Image(BaseModel):    
    data: Optional[Union[str, bytes]]
    url: Optional[str]
    filepath: Optional[str]
    media_type: Literal["image/jpeg", "image/png", "image/gif", "image/webp"] = "image/jpeg"
    format: Literal["base64", "url", "filepath"] = "base64"

class ToolCall(BaseModel):
    tool_call_id: str
    tool_name: str
    arguments: Optional[Mapping[str, Any]]
    output: Optional[Any] = None
    type: str = "function"


class Message(BaseModel):
    role: Literal['user', 'assistant', 'system', 'tool']
    content: Optional[str] = None
    images: Optional[list[Image]] = None
    tool_calls: Optional[list[ToolCall]] = None
    response_object: Optional[Any] = None
    # response_info: Optional[ResponseInfo]


ToolParameterType = Literal["object", "array", "string", "integer", "number", "boolean", "null"]
ToolValPyTypes = Union[str, int, float, bool, None, list, dict]
ToolDict = dict[str, ToolValPyTypes]


class ToolParameter(BaseModel):
    type: ToolParameterType = "string"
    name: Optional[str] = None
    description: Optional[str] = None
    enum: Optional[list[ToolValPyTypes]] = None
    required: bool = False
    # enum: list[str] = Field(default_factory=list)
    

    def to_dict(self) -> ToolDict:
        param_dict: ToolDict = {"type": self.type}
        if self.description:
            param_dict["description"] = self.description
        if self.enum:
            param_dict["enum"] = self.enum
        return param_dict
    
class StringToolParameter(ToolParameter):
    type: Literal["string"] = "string"

class NumberToolParameter(ToolParameter):
    type: Literal["number"] = "number"

class IntegerToolParameter(ToolParameter):
    type: Literal["integer"] = "integer"

class BooleanToolParameter(ToolParameter):
    type: Literal["boolean"] = "boolean"

class NullToolParameter(ToolParameter):
    type: Literal["null"] = "null"

class ArrayToolParameter(ToolParameter):
    type: Literal["array"] = "array"
    items: ToolParameter
    
    def to_dict(self) -> ToolDict:
        return {
            **ToolParameter.to_dict(self),
            "items": self.items.to_dict() 
        }
    
class ObjectToolParameter(ToolParameter):
    type: Literal["object"] = "object"
    properties: list[ToolParameter]
    additionalProperties: bool = False
    
    def to_dict(self) -> ToolDict:
        properties = {}
        required = []
        for prop in self.properties:
            properties[prop.name] = prop.to_dict()
            if prop.required:
                required.append(prop.name)

        return { 
            **ToolParameter.to_dict(self),
            "properties": properties,
            "required": required,
            "additionalProperties": self.additionalProperties
        }
    
class AnyOfToolParameter(ToolParameter):
    type: Literal["anyOf"] = "anyOf"
    anyOf: list[ToolParameter]

    def to_dict(self) -> ToolDict:
        return {
            "anyOf": [param.to_dict() for param in self.anyOf]
        }

    
class Tool(BaseModel):
    type: str
    name: str

    def to_dict(self) -> dict:
        return {
            "type": self.type,
            # self.type: {
            #     "name": self.name
            # }
        }

class FunctionTool(Tool):
    type: Literal["function"] = "function"
    description: str
    parameters: Union[ObjectToolParameter, ArrayToolParameter, ToolParameter]
    strict: bool = True
    callable: Optional[Callable] = None

    def to_dict(self):
        return {
            "type": self.type,
            self.type: {
                "name": self.name,
                "description": self.description,
                "parameters": self.parameters.to_dict(),
                "strict": self.strict
            },            
        }

class CodeInterpreterTool(Tool):
    type: Literal["code_interpreter"] = "code_interpreter"
    name: str = "code_interpreter"

class FileSearchTool(Tool):
    type: Literal["file_search"] = "file_search"
    name: str = "file_search"


class EvalTypeParameters(BaseModel):
    name: str
    system_prompt: str = "Your role is to evaluate the content using the provided tool(s)." 
    examples: Optional[list[Union[Message, dict[Literal["input", "response"], Any]]]] = None   
    tools: Optional[list[Union[Tool, str]]] = None
    tool_choice: Optional[Union[str, list[str]]] = None
    return_on: Optional[Union[Literal["content", "tool_call", "message"], str, list[str]]] = None
    return_as: Literal["messages", "last_message", "last_content_or_tool_call_args", "last_content_or_all_tool_call_args"] = "last_content_or_tool_call_args"

    
    response_format: Optional[Union[str, dict[str, str]]] = None
    enforce_tool_choice: bool = True
    tool_choice_error_retries: int = 3




# © 2024 Lucas Faudman.
# Licensed under the MIT License (see LICENSE for details).
# For commercial use, see LICENSE for additional terms.
from setuptools import setup, find_namespace_packages
from setuptools.command.build_ext import build_ext

EXT_MODULES = []
# try:
#     from mypyc.build import mypycify

#     EXT_MODULES.extend(
#         mypycify(
#             [
#                 "src/unifai/unifai.py",
#                 "src/unifai/concurrent_executor.py",
#                 "src/unifai/decompiler.py",
#                 "src/unifai/secret_scanner.py",
#             ]
#         )
#     )
# except Exception as e:
#     print(f"Failed to compile with mypyc: {e}")

setup(
    name="unifai",
    version="0.0.1",
    use_scm_version=True,
    setup_requires=["setuptools>=42", "setuptools_scm>=8", "wheel"],
    description="Unify AI clients into a single interface with enhanced Tool Calling support.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Lucas Faudman",
    author_email="lucasfaudman@gmail.com",
    url="https://github.com/LucasFaudman/unifai.git",
    packages=find_namespace_packages(where="src", exclude=["tests*"]),
    package_dir={"": "src"},
    package_data={
        "": ["LICENSE"],
    },
    include_package_data=True,
    exclude_package_data={"": [".gitignore", ".pre-commit-config.yaml"]},
    install_requires=["pydantic"],
    ext_modules=EXT_MODULES,
    cmdclass={"build_ext": build_ext},
    extras_require={
        "dev": ["pytest"],
        "openai": ["openai"],
        "ollama": ["ollama"],
        "anthropic": ["anthropic"],
    },
    entry_points={
        "console_scripts": [],
    },
    python_requires=">=3.10",
    license="LICENSE",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="",
    project_urls={
        "Homepage": "https://github.com/LucasFaudman/unifai.git",
        "Repository": "https://github.com/LucasFaudman/unifai.git",
    },
)





from typing import Optional, Union, Any, Literal, Mapping, Type, Callable, Collection, Sequence, Type, List, Dict, Tuple

from json import dumps as json_dumps
from .baseaiclientwrapper import BaseAIClientWrapper

from ._types import Message, Tool, ToolCall, EvalTypeParameters
from ._convert_types import tool_from_dict, stringify_content, make_few_shot_prompt

AIProvider = Literal["anthropic", "openai", "ollama"]

ToolInput = Union[Tool, dict[str, Any], str]
EvalTypeParametersInput = Union[EvalTypeParameters, dict[str, Any]]

class UnifAIClient:
    TOOLS: list[ToolInput] = []
    TOOL_CALLABLES: dict[str, Callable] = {}
    EVAL_TYPES: list[EvalTypeParametersInput] = []


    def import_client_wrapper(self, provider: AIProvider) -> Type[BaseAIClientWrapper]:
        match provider:
            case "anthropic":
                from .anthropic_wrapper import AnthropicWrapper
                return AnthropicWrapper
            case "openai":
                from .openai_wrapper import OpenAIWrapper
                return OpenAIWrapper
            case "ollama":
                from .ollama_wrapper import OllamaWrapper
                return OllamaWrapper

    
    def __init__(self, 
                 provider_client_kwargs: Optional[dict[AIProvider, dict[str, Any]]] = None,
                 tools: Optional[list[ToolInput]] = None,
                 tool_callables: Optional[dict[str, Callable]] = None,
                 eval_types: Optional[list[EvalTypeParametersInput]] = None

                 ):
        self.provider_client_kwargs = provider_client_kwargs if provider_client_kwargs is not None else {}
        self.providers = list(self.provider_client_kwargs.keys())
        self.default_provider: AIProvider = self.providers[0] if len(self.providers) > 0 else "openai"
        
        self._clients: dict[AIProvider, BaseAIClientWrapper] = {}
        self.tools: dict[str, Tool] = {}
        self.tool_callables: dict[str, Callable] = {}
        self.eval_types: dict[str, EvalTypeParameters] = {}
        
        self.add_tools(tools or self.TOOLS)
        self.add_tool_callables(tool_callables)
        self.add_eval_types(eval_types or self.EVAL_TYPES)
    

    def add_tools(self, tools: Optional[list[ToolInput]]):
        if not tools: return

        for tool in self.standardize_tools(tools):
            self.tools[tool.name] = tool
            if (tool_callable := getattr(tool, "callable", None)) is not None:
                self.tool_callables[tool.name] = tool_callable

    def add_tool_callables(self, tool_callables: Optional[dict[str, Callable]]):
        if not tool_callables: return
        self.tool_callables.update(tool_callables)


    def add_eval_types(self, eval_types: Optional[list[EvalTypeParametersInput]]):
        if not eval_types: return
        self.eval_types.update(self.standardize_eval_types(eval_types))

        
    def init_client(self, provider: AIProvider, **client_kwargs) -> BaseAIClientWrapper:
        client_kwargs = {**self.provider_client_kwargs[provider], **client_kwargs}
        self._clients[provider] = self.import_client_wrapper(provider)(**client_kwargs)
        return self._clients[provider]
    
    def get_client(self, provider: Optional[AIProvider] = None) -> BaseAIClientWrapper:
        provider = provider or self.default_provider
        if provider not in self._clients:
            return self.init_client(provider)
        return self._clients[provider]

    # List Models
    def list_models(self, provider: Optional[AIProvider] = None) -> list[str]:
        return self.get_client(provider).list_models()

    def standardize_eval_types(self, eval_types: Sequence[EvalTypeParametersInput]) -> dict[str, EvalTypeParameters]:
        std_eval_types = {}
        for eval_type in eval_types:
            if isinstance(eval_type, EvalTypeParameters):
                std_eval_types[eval_type.name] = eval_type
            elif isinstance(eval_type, dict):
                std_eval_types[eval_type['eval_type']] = EvalTypeParameters(**eval_type)
            else:
                raise ValueError(f"Invalid eval_type type: {type(eval_type)}")
        return std_eval_types


    def standardize_messages(self, messages: Sequence[Union[Message, str, dict[str, Any]]]) -> list[Message]:
        std_messages = []
        for message in messages:
            if isinstance(message, Message):
                std_messages.append(message)
            elif isinstance(message, str):
                std_messages.append(Message(role="user", content=message))
            elif isinstance(message, dict):
                std_messages.append(Message(**message))
            else:
                raise ValueError(f"Invalid message type: {type(message)}")        
        return std_messages
    
    
    def standardize_tools(self, tools: Sequence[ToolInput]) -> list[Tool]:
        std_tools = []
        for tool in tools:
            if isinstance(tool, Tool):
                std_tools.append(tool)
            elif isinstance(tool, dict):
                std_tools.append(tool_from_dict(tool))
            elif isinstance(tool, str):
                if std_tool := self.tools.get(tool):
                    std_tools.append(std_tool)
                else:
                    raise ValueError(f"Tool '{tool}' not found in tools")
            else:
                raise ValueError(f"Invalid tool type: {type(tool)}")
        
        return std_tools
    
    def standardize_tool_choice(self, tool_choice: Union[Literal["auto", "required", "none"], Tool, str, dict]) -> str:
        if isinstance(tool_choice, Tool):
            return tool_choice.name
        if isinstance(tool_choice, dict):
            tool_type = tool_choice['type']
            return tool_choice[tool_type]['name']
        
        # tool_choice is a string tool_name or Literal value "auto", "required", or "none"
        return tool_choice

    def filter_tools_by_tool_choice(self, tools: list[Tool], tool_choice: str) -> list[Tool]:
        if tool_choice == "auto" or tool_choice == "required":
            return tools
        if tool_choice == "none":
            return []
        return [tool for tool in tools if tool.name == tool_choice]

    
    def check_tool_choice_obeyed(self, tool_choice: str, tool_calls: Optional[list[ToolCall]]) -> bool:
        if tool_calls:
            tool_names = [tool_call.tool_name for tool_call in tool_calls]
            if (
                # tools were called but tool choice is none
                tool_choice == 'none'
                # the correct tool was not called and tool choice is not "required" (required=any one or more tools must be called) 
                or (tool_choice != 'required' and tool_choice not in tool_names)
                ):
                print(f"Tools called and tool_choice={tool_choice} NOT OBEYED")
                return False
        elif tool_choice != 'none':
            print(f"Tools NOT called and tool_choice={tool_choice} NOT OBEYED")
            return False 
        
        print(f"tool_choice={tool_choice} OBEYED")
        return True       


    def do_tool_calls(self, tool_calls: list[ToolCall]) -> list[ToolCall]:
        for tool_call in tool_calls:
            if tool_callable := self.tool_callables.get(tool_call.tool_name):
                # TODO catch ToolCallError
                tool_call.output = tool_callable(**(tool_call.arguments or {}))
            else:
                raise ValueError(f"Tool '{tool_call.tool_name}' callable not found")
        
        return tool_calls

    # Chat
    def chat(
            self,
            messages: Sequence[Union[Message, str, dict[str, Any]]],
            provider: Optional[AIProvider] = None,            
            model: Optional[str] = None,
            system_prompt: Optional[str] = None,             
            tools: Optional[Sequence[ToolInput]] = None,
            tool_choice: Optional[Union[Literal["auto", "required", "none"], Tool, str, dict, Sequence[Union[Tool, str, dict]]]] = None,
            response_format: Optional[Union[str, dict[str, str]]] = None,

            return_on: Optional[Union[Literal["content", "tool_call", "message"], str, Collection[str]]] = "content",
            enforce_tool_choice: bool = False,
            tool_choice_error_retries: int = 3,
            
            **kwargs
            ) -> list[Message]:
        
        # import and init client
        client = self.get_client(provider)
        
        # standardize inputs and prep copies for client in its format
        # (Note to self: 2 copies are stored to prevent converting back and forth between formats on each iteration.
        # this choice is at the cost of memory but is done to prevent unnecessary conversions 
        # and allow for easier debugging and error handling.
        # May revert back to optimizing for memory later if needed.)
        model = model or client.default_model
        std_messages = self.standardize_messages(messages)
        # client_messages = [client.prep_input_message(message) for message in std_messages]
        client_messages, system_prompt = client.prep_input_messages_and_system_prompt(std_messages, system_prompt)

        if tools:
            std_tools = self.standardize_tools(tools)
            client_tools = [client.prep_input_tool(tool) for tool in std_tools]
        else:
            std_tools = client_tools = None

        if tool_choice:
            if isinstance(tool_choice, Sequence) and not isinstance(tool_choice, str):
                std_tool_choice = self.standardize_tool_choice(tool_choice[0])
                std_tool_choice_queue = [self.standardize_tool_choice(tool_choice) for tool_choice in tool_choice[1:]]
            else:
                std_tool_choice = self.standardize_tool_choice(tool_choice)
                std_tool_choice_queue = None
            client_tool_choice = client.prep_input_tool_choice(std_tool_choice)
        else:
            std_tool_choice = std_tool_choice_queue = client_tool_choice = None

        # if tools:
        #     std_tools = self.standardize_tools(tools)
        #     # client_tools = [client.prep_input_tool(tool) for tool in std_tools]
        #     std_tools_filtered = self.filter_tools_by_tool_choice(std_tools, std_tool_choice) if std_tool_choice else std_tools
        #     client_tools = [client.prep_input_tool(tool) for tool in std_tools_filtered]
        # else:
        #     std_tools = client_tools = None

        if response_format:
            client_response_format = client.prep_input_response_format(response_format)
        else:
            client_response_format = None
        
        # Chat and ToolCall handling loop
        while (response := client.chat(
                messages=client_messages,                 
                model=model, 
                system_prompt=system_prompt,
                tools=client_tools, 
                tool_choice=client_tool_choice,
                response_format=client_response_format, 
                **kwargs
            )):

            # TODO check if response is an APIError

            std_message, client_message = client.extract_std_and_client_messages(response)
            print("std_message:", std_message)

            # Enforce Tool Choice: Check if tool choice is obeyed
            # auto:  
            if enforce_tool_choice and std_tool_choice != 'auto' and std_tool_choice is not None:
                if self.check_tool_choice_obeyed(std_tool_choice, std_message.tool_calls):
                    # TODO implement tool choice sequence
                    if std_tool_choice_queue and std_tools:
                        # Update std_tools and client_tools with next tool choice
                        std_tool_choice = std_tool_choice_queue.pop(0)
                        client_tool_choice = client.prep_input_tool_choice(std_tool_choice)

                        # # # Filter tools by tool choice            
                        # std_tools_filtered = self.filter_tools_by_tool_choice(std_tools, std_tool_choice)
                        # client_tools = [client.prep_input_tool(tool) for tool in std_tools_filtered]
                    else:
                        std_tool_choice = client_tool_choice = None # set to auto for next iteration
                        
                elif tool_choice_error_retries > 0:
                    tool_choice_error_retries -= 1
                    # Continue to next iteration without updating messages (retry)
                    continue
                else:
                    print("Tool choice error retries exceeded")
                    raise ValueError("Tool choice error retries exceeded")

            # Update messages with assistant message
            std_messages.append(std_message)
            client_messages.append(client_message)
            
            if return_on == "message":
                print("returning on message")
                break
            
            if tool_calls := std_message.tool_calls:
                # Return on tool_call before processing messages
                if any(
                    return_on == "tool_call" # return_on is a string "tool_call"
                    or tool_call.tool_name == return_on # return_on is a tool name
                    or (isinstance(return_on, Collection) and tool_call.tool_name in return_on) # return_on is a collection of tool names
                    for tool_call in tool_calls):
                    print("returning on tool call", return_on)
                    break

                tool_calls = self.do_tool_calls(tool_calls)
                for std_tool_message in client.split_tool_call_outputs_into_messages(tool_calls):
                    std_messages.append(std_tool_message)
                    client_messages.append(client.prep_input_message(std_tool_message))

                # for tool_call in tool_calls:
                #     std_tool_output_message = self.do_tool_call(tool_call)
                #     client_tool_output_message = client.prep_input_message(std_tool_output_message)
                #     # Update messages with tool outputs
                #     std_messages.append(std_tool_output_message)
                #     client_messages.append(client_tool_output_message)
                
                # Process messages after submitting tool outputs
                continue

            print("Returning on content:", std_message.content)
            break

        return std_messages

    
    def evaluate(self, 
                 eval_type: str, 
                 content: Any, 
                 provider: Optional[AIProvider] = None,
                 model: Optional[str] = None,
                 **kwargs
                 ) -> Any:
        
        if (eval_type_parameters := self.eval_types.get(eval_type)) is None:
            raise ValueError(f"Eval type '{eval_type}' not found in eval_types")

        if isinstance(eval_type_parameters.tool_choice, str):
            last_tool_choice = eval_type_parameters.tool_choice
        elif eval_type_parameters.tool_choice:
            last_tool_choice = eval_type_parameters.tool_choice[-1]
        else:
            last_tool_choice = None

        return_on = eval_type_parameters.return_on or last_tool_choice or "content"

        input_messages = make_few_shot_prompt(
            system_prompt=eval_type_parameters.system_prompt,
            examples=eval_type_parameters.examples,
            content=content
        )

        messages = self.chat(
            messages=input_messages,
            provider=provider,
            model=model,
            tools=eval_type_parameters.tools,
            tool_choice=eval_type_parameters.tool_choice,
            response_format=eval_type_parameters.response_format,
            return_on=return_on,
            **kwargs
        )
        
        return_as = eval_type_parameters.return_as
        if return_as == "messages":
            return messages
        
        last_message = messages[-1]
        if return_as == "last_message":
            return last_message
        
        if last_message.content:
            return last_message.content
        
        if last_message.tool_calls:                
            if return_as == "last_content_or_all_tool_call_args":
                return [tool_call.arguments for tool_call in last_message.tool_calls]
            
            # default return_as == "last_content_or_tool_call_args"
            return last_message.tool_calls[-1].arguments
        
        raise ValueError("No content or tool call arguments to return")
            
